from typing import TypeVar, Generic
from abc import ABC, ABCMeta, abstractmethod


class ChangelogEntryError(Exception):
    
    def __init__(self, *args):
        super().__init__(*args)


class ChangelogEntry(ABC):
    """
    Represents on entry in the changelog.
    
    Note that while the id, change time and user who modified, are part of an entry, they are not saved in this object
    because this object is also used for entries which are yet to be written to the database.
    """
    __metaclass__ = ABCMeta
    
    def __init__(self,
                 subject: "ChangelogEntrySubject",
                 object_id: int):
        super().__init__()
        self.subject = subject
        self.object_id = object_id
    
    def add_parameters_to_api_json(self, entry_json: dict):
        entry_json["object_type"] = self.subject.get_object_class().api_id
        entry_json["object_id"] = self.object_id


EntryTypeGeneric = TypeVar("EntryTypeGeneric", bound=ChangelogEntry)


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ChangelogEntrySubject(Generic[EntryTypeGeneric], ABC):
    """
    The subject for a :class:`ChangelogEntry`. A subject is bound to a specific object class and represents 'something'
    that can be 'modified'.
    For example, ObjectField is an entry subject since it represents a single value which can be modified. All such
    entries are of type ModificationEntry.
    
    For every log version the following exists:
    
    Every entry subject has one id which uniquely identifies that subject in it's object class. And every entry subject
    has one codec, which can decode and possibly encode a changelog entry of that subject (from that version)
    """
    __metaclass__ = ABCMeta
    
    @abstractmethod
    def get_object_class(self) -> "ObjectClass":
        pass  # pragma: no cover


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class SimpleChangelogEntrySubject(ChangelogEntrySubject):
    
    def __init__(self, _object_class: "ObjectClass"):
        super().__init__()
        self._object_class = _object_class
    
    def get_object_class(self) -> "ObjectClass":
        return self._object_class


EntrySubjectGeneric = TypeVar("EntrySubjectGeneric", bound=ChangelogEntrySubject)


class ChangelogEntryCodec(ABC, Generic[EntrySubjectGeneric]):
    """
    The codec which can decode (and maybe encode) entries of one :subject:`ChangelogEntrySubject` or a range of entry subjects.
    """
    
    @abstractmethod
    def decode_entry(self,
                     subject: EntrySubjectGeneric,
                     object_id: int,
                     param1: str or None,
                     param2: str) -> ChangelogEntry:
        pass  # pragma: no cover
    
    def supports_encoding(self) -> bool:
        return False  # pragma: no cover
    
    @abstractmethod
    def encode_entry(self, entry: ChangelogEntry) -> tuple[int, str or None, str]:
        pass  # pragma: no cover


class _VersionedEntrySubject:
    
    def __init__(self,
                 subject: ChangelogEntrySubject,
                 object_id: str,
                 id: str,
                 codec: ChangelogEntryCodec):
        super().__init__()
        self.subject = subject
        self.object_id = object_id
        self.id = id
        self.codec = codec
    
    def decode(self,
               object_id: int,
               param1: str or None,
               param2: str) -> ChangelogEntry:
        return self.codec.decode_entry(self.subject, object_id, param1, param2)
    
    def encode(self, entry: ChangelogEntry) -> tuple[int, str or None, str]:
        if entry.subject != self.subject:
            raise RuntimeError("Entry is for different subject")  # pragma: no cover
        return self.codec.encode_entry(entry)


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class _ChangelogVersion:
    
    def __init__(self,
                 id: int,
                 object_id_by_class: dict["ObjectClass", str],
                 object_class_by_id: dict[str, "ObjectClass"],
                 versioned_subject_by_subject: dict[ChangelogEntrySubject, _VersionedEntrySubject],
                 versioned_subject_by_id_by_object: dict["ObjectClass", dict[str, _VersionedEntrySubject]]):
        super().__init__()
        self.id = id
        self.object_id_by_class = object_id_by_class
        self.object_class_by_id = object_class_by_id
        self.versioned_subject_by_subject = versioned_subject_by_subject
        self.versioned_subject_by_id_by_object = versioned_subject_by_id_by_object
    
    def decode_entry(self,
                     object_class_id: str,
                     object_id: int,
                     subject_id: str,
                     param1: str or None,
                     param2: str) -> ChangelogEntry:
        if object_class_id not in self.object_class_by_id:
            raise ChangelogEntryError(f"Unknown object class {object_class_id} for version v{self.id}")
        object_class = self.object_class_by_id[object_class_id]
        
        if subject_id not in self.versioned_subject_by_id_by_object[object_class]:
            raise ChangelogEntryError(f"Unknown subject {subject_id} for object {object_class_id} in version v{self.id}")
        
        return self.versioned_subject_by_id_by_object[object_class][subject_id].decode(object_id, param1, param2)
        
    def encode_entry(self, entry: ChangelogEntry) -> tuple[int, str or None, str]:
        if entry.subject not in self.versioned_subject_by_subject:
            raise ChangelogEntryError(f"Unknown subject {entry.subject} in version v{self.id}")
        return self.versioned_subject_by_subject[entry.subject].encode(entry)


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ChangelogVersionManager:
    """
    This manages the ids and codecs of all :class:`ChangelogEntryClass` for all versions
    """
    def __init__(self,
                 versions_by_id: dict[int, _ChangelogVersion],
                 current_version: int):
        super().__init__()
        self._versions_by_id = versions_by_id
        self._current_version = current_version
    
    def get_current_version_id(self):
        return self._current_version
    
    def get_all_versions_object_id(self, object_class: "ObjectClass") -> list[tuple[int, str]]:
        ids: list[tuple[int, str]] = []
        for version_id, version in self._versions_by_id.items():
            object_id = version.object_id_by_class.get(object_class, None)
            if object_id is not None:
                ids.append((version_id, object_id))
        return ids
    
    def get_all_versions_subject_id(self, subject: ChangelogEntrySubject) -> list[tuple[int, str, str]]:
        ids: list[tuple[int, str, str]] = []
        for version_id, version in self._versions_by_id.items():
            versioned_subject = version.versioned_subject_by_subject.get(subject, None)
            if versioned_subject is not None:
                ids.append((version_id, versioned_subject.object_id, versioned_subject.id))
        return ids
    
    def _get_versioned_subject(self, subject: ChangelogEntrySubject, version: int) -> _VersionedEntrySubject or None:
        if version not in self._versions_by_id:
            return None
        version = self._versions_by_id[version]
        if subject not in version.versioned_subject_by_subject:
            return None
        return version.versioned_subject_by_subject[subject]

    def get_subject_id(self, subject: ChangelogEntrySubject, version: int) -> tuple[str, str] or None:
        versioned = self._get_versioned_subject(subject, version)
        if versioned is None:
            return None
        return versioned.object_id, versioned.id
    
    def can_encode_for(self, subject: ChangelogEntrySubject, version: int):
        versioned = self._get_versioned_subject(subject, version)
        return versioned is not None and versioned.codec.supports_encoding()
    
    def decode_entry(self,
                     log_version: int,
                     object_class_id: str,
                     object_id: int,
                     subject_id: str,
                     param1: str or None,
                     param2: str) -> ChangelogEntry:
        if log_version not in self._versions_by_id:
            raise ChangelogEntryError(f"Unknown log version v{log_version}")
        return self._versions_by_id[log_version].decode_entry(object_class_id, object_id, subject_id, param1, param2)
    
    def encode_entry(self,
                     log_version: int,
                     entry: ChangelogEntry) -> tuple[int, str or None, str]:
        if log_version not in self._versions_by_id:
            raise ChangelogEntryError(f"Unknown log version v{log_version}")
        return self._versions_by_id[log_version].encode_entry(entry)


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ChangelogVersionBuilder:
    """
    Used to build a :class:`ChangelogVersionManager`
    """
    
    def __init__(self):
        super().__init__()
        self._versions: dict[int, _ChangelogVersion] = {}
        
        self._current_version = -1
        self._current_object_class_by_id: dict[str, "ObjectClass"] = {}
        self._current_object_id_by_class: dict["ObjectClass", str] = {}
        self._current_subject_by_id_by_object_class: dict["ObjectClass", dict[str, tuple[ChangelogEntrySubject, ChangelogEntryCodec]]] = {}
        self._current_subject_id_by_subject: dict[ChangelogEntrySubject, str] = {}
        
        self._highest_version = -1
    
    def finish_version(self):
        if self._current_version == -1:
            raise RuntimeError("No version is currently being build")  # pragma: no cover
        
        object_id_by_class: dict["ObjectClass", str] = self._current_object_id_by_class
        object_class_by_id: dict[str, "ObjectClass"] = self._current_object_class_by_id
        versioned_subject_by_subject: dict[ChangelogEntrySubject, _VersionedEntrySubject] = {}
        versioned_subject_by_id_by_object: dict["ObjectClass", dict[str, _VersionedEntrySubject]] = {}
        
        for object_class, subjects_dict in self._current_subject_by_id_by_object_class.items():
            object_id = self._current_object_id_by_class[object_class]
            versioned_subject_by_id_by_object[object_class] = {}
            for subject_id, (subject, codec) in subjects_dict.items():
                versioned_subject = _VersionedEntrySubject(
                    subject,
                    object_id,
                    subject_id,
                    codec
                )
                versioned_subject_by_subject[subject] = versioned_subject
                versioned_subject_by_id_by_object[object_class][subject_id] = versioned_subject
        
        new_version = _ChangelogVersion(
            self._current_version,
            object_id_by_class,
            object_class_by_id,
            versioned_subject_by_subject,
            versioned_subject_by_id_by_object
        )
        self._versions[self._current_version] = new_version
        self._current_version = -1
        self._current_object_class_by_id = {}
        self._current_object_id_by_class = {}
        self._current_subject_by_id_by_object_class = {}
        self._current_subject_id_by_subject = {}
    
    def start_version(self, id: int, depends_on: int or None = None):
        """
        :param id: ID for this version. Must not be negative
        :param depends_on: ID of an already finished version on which this version depends. All entries from that
                           version are copied for the new version, and changes can then be made.
        """
        if self._current_version != -1:
            raise RuntimeError("Already building a version. Finish previous version with finish_version()")  # pragma: no cover
        if id < 0:
            raise RuntimeError("Version id cannot be negative")  # pragma: no cover
        if depends_on is not None:
            if depends_on not in self._versions:
                raise RuntimeError("Unknown depends_on version")  # pragma: no cover
            depends_version = self._versions[depends_on]
            self._current_object_class_by_id = depends_version.object_class_by_id
            self._current_object_id_by_class = depends_version.object_id_by_class
            for clazz in self._current_object_id_by_class.keys():
                self._current_subject_by_id_by_object_class[clazz] = {}
            for subject, _versioned_subject in depends_version.versioned_subject_by_subject.items():
                self._current_subject_by_id_by_object_class[subject.get_object_class()][_versioned_subject.id] = \
                    (subject, _versioned_subject.codec)
                self._current_subject_id_by_subject[subject] = _versioned_subject.id
        
        self._current_version = id
        self._highest_version = max(self._highest_version, id)
    
    def set_object_class_id(self, object_class: "ObjectClass", id: str):
        if id in self._current_object_class_by_id and self._current_object_class_by_id[id] != object_class:
            raise RuntimeError(f"That id is already assigned to class {self._current_object_class_by_id[id]}")  # pragma: no cover
        if object_class in self._current_object_id_by_class:
            self._current_object_class_by_id.pop(self._current_object_id_by_class[object_class])
        self._current_object_class_by_id[id] = object_class
        self._current_object_id_by_class[object_class] = id
        if object_class not in self._current_subject_by_id_by_object_class:
            self._current_subject_by_id_by_object_class[object_class] = {}
    
    def set_subject_id_and_codec(self, subject: ChangelogEntrySubject, id: str, codec: ChangelogEntryCodec):
        clazz = subject.get_object_class()
        if clazz not in self._current_object_id_by_class:
            raise RuntimeError("Object class of subject is not known")  # pragma: no cover
        if id in self._current_subject_by_id_by_object_class[clazz]:  # pragma: no cover
            raise RuntimeError(f"That id is already assigned to subject "
                               f"{self._current_subject_by_id_by_object_class[clazz][id][0]} (Ids must be unique for one "
                               f"object class)")
        if subject in self._current_subject_id_by_subject:
            self._current_subject_by_id_by_object_class[clazz].pop(self._current_subject_id_by_subject[subject])
        self._current_subject_by_id_by_object_class[clazz][id] = (subject, codec)
        self._current_subject_id_by_subject[subject] = id
    
    def build(self) -> ChangelogVersionManager:
        if self._current_version != -1:
            raise RuntimeError("Currently building a version. Finish version with finish_version()")  # pragma: no cover
        if self._highest_version == -1:
            raise RuntimeError("No versions added")  # pragma: no cover
        return ChangelogVersionManager(
            self._versions.copy(),
            self._highest_version
        )
