import json

from api.objects.changelog import ChangelogVersionBuilder, ChangelogEntryCodec, ChangelogEntryError, SimpleChangelogEntrySubject
from api.objects.changelog_entries import ModificationEntry, DeletionEntry, CreationEntry
from api.objects.field import ObjectField


class ModificationEntryCodec(ChangelogEntryCodec[ObjectField]):
    
    def decode_entry(self,
                     subject: ObjectField,
                     object_id: int,
                     param1: str or None,
                     param2: str) -> ModificationEntry:
        try:
            value_old_json = None
            if param1 is not None:
                value_old_json = json.loads(param1)
            value_new_json = json.loads(param2)
        except ValueError as e:
            raise ChangelogEntryError("Unable to decode json") from e
        return ModificationEntry(
            subject,
            object_id,
            value_old_json,
            value_new_json,
            param1 is None
        )
    
    def supports_encoding(self) -> bool:
        return True
    
    def encode_entry(self, entry: ModificationEntry) -> tuple[int, str or None, str]:
        return (entry.object_id,
                None if entry.is_creation else json.dumps(entry.value_old_json),
                json.dumps(entry.value_new_json))


class DeletionEntryCodec(ChangelogEntryCodec[SimpleChangelogEntrySubject]):
    
    def decode_entry(self,
                     subject: SimpleChangelogEntrySubject,
                     object_id: int,
                     param1: str or None,
                     param2: str) -> DeletionEntry:
        return DeletionEntry(
            subject,
            object_id,
            bool(json.loads(param2))
        )
    
    def supports_encoding(self) -> bool:
        return True
    
    def encode_entry(self, entry: DeletionEntry) -> tuple[int, str or None, str]:
        return entry.object_id, json.dumps(not entry.is_now_deleted), json.dumps(entry.is_now_deleted)


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class CreationEntryCodec(ChangelogEntryCodec[SimpleChangelogEntrySubject]):
    
    def decode_entry(self,
                     subject: SimpleChangelogEntrySubject,
                     object_id: int,
                     param1: str or None,
                     param2: str) -> CreationEntry:
        try:
            data = json.loads(param2)
            variant: str or None = None
            if "variant" in data:
                variant = data["variant"]
            
            parent_type: "ObjectClass" or None = None
            parent_id: int or None = None
            if "parent_type" in data:
                from api.objects.objects import object_classes_by_id
                type_id: str = data["parent_type"]
                if type_id not in object_classes_by_id:
                    raise ChangelogEntryError("Unknown object type " + type_id)
                parent_type = object_classes_by_id[type_id]
                parent_id = data["parent_id"]
        except (ValueError, TypeError) as e:
            raise ChangelogEntryError(f"Invalid creation entry data: {e}") from e
        
        return CreationEntry(
            subject,
            object_id,
            variant,
            parent_type,
            parent_id
        )
    
    def supports_encoding(self) -> bool:
        return True
    
    def encode_entry(self, entry: CreationEntry) -> tuple[int, str or None, str]:
        data = {}
        if entry.variant is not None:
            data["variant"] = entry.variant
        if entry.parent_type is not None:
            data["parent_type"] = entry.parent_type.api_id
            data["parent_id"] = entry.parent_id
        return (entry.object_id,
                None,
                json.dumps(data))


_MODIFICATION_CODEC = ModificationEntryCodec()
_DELETION_CODEC = DeletionEntryCodec()
_CREATION_CODEC = CreationEntryCodec()


def changelog_v2_build(builder: ChangelogVersionBuilder):
    from api.objects.objects import object_classes_by_id
    
    builder.start_version(2)
    for class_id, clazz in object_classes_by_id.items():
        builder.set_object_class_id(clazz, class_id)
        for field_id in clazz.get_all_field_ids():
            field = clazz.get_field(field_id)
            builder.set_subject_id_and_codec(field, field_id, _MODIFICATION_CODEC)
        builder.set_subject_id_and_codec(clazz.get_deletion_subject(), "is_deleted", _DELETION_CODEC)
        builder.set_subject_id_and_codec(clazz.get_creation_subject(), "creation", _CREATION_CODEC)
    builder.finish_version()
