from datetime import datetime
from typing import Callable
import math

from api.database import *
from api.miscellaneous import *
from api.objects.changelog import ChangelogVersionManager, ChangelogVersionBuilder, ChangelogEntry, ChangelogEntryError, \
    ChangelogEntrySubject
from api.objects.changelog_v1 import changelog_v1_build
from api.objects.changelog_v2 import changelog_v2_build
from api.objects.object import ObjectClass

CHANGELOG_MIN_PAGE_SIZE = 10
CHANGELOG_MAX_PAGE_SIZE = 500


def _build_changelog_manager() -> ChangelogVersionManager:
    builder = ChangelogVersionBuilder()
    changelog_v1_build(builder)
    changelog_v2_build(builder)
    return builder.build()


_VERSION_MANAGER = _build_changelog_manager()
_LEGACY_VERSION = 1


def _changelog_entry_db_to_json(entry_db: DbResultRow) -> dict:
    id = entry_db["id"]
    change_time: datetime = entry_db["when"]
    modifying_user_id = entry_db["who"]
    table = str(entry_db["table"])
    field = str(entry_db["field"])
    try:
        id_value = int(entry_db["id_value"])
    except ValueError:
        id_value = -1  # There are a few entries in the db which contain a string
    value_old = entry_db["value_old"]
    value_new = str(entry_db["value_new"])
    
    entry_json = {
        "id": id,
        "change_time": change_time.strftime(API_DATETIME_FORMAT),
        "modifying_user_id": modifying_user_id
    }
    try:
        _VERSION_MANAGER.decode_entry(
            entry_db["log_version"],
            table,
            id_value,
            field,
            value_old,
            value_new
        ).add_parameters_to_api_json(entry_json)
    except ChangelogEntryError as e:
        if DEBUG_ENABLED:
            print("Warning: Exception while decoding changelog entry (This may happen for old entries): " + str(e))
        entry_json["type"] = "unknown"
        entry_json["object_id"] = id_value
        entry_json["unknown_type"] = table
        entry_json["unknown_field"] = field
        entry_json["old_value"] = value_old
        entry_json["new_value"] = value_new
    return entry_json


def get_changelog(entries_per_page: int,
                  page: int,
                  filter_object_class: ObjectClass or None = None,
                  filter_object_id: int or None = None,
                  filter_subject: ChangelogEntrySubject or None = None) -> tuple[int, JsonTypes]:
    if entries_per_page < CHANGELOG_MIN_PAGE_SIZE or entries_per_page > CHANGELOG_MAX_PAGE_SIZE or page < 0:
        raise ValueError("Invalid page size or page number")  # pragma: no cover
    
    if filter_subject is not None and filter_object_class is not None and filter_subject.get_object_class() != filter_object_class:
        raise RuntimeError("If subject is set, object class must be class of subject")  # pragma: no cover
    
    clauses = []
    all_values = []
    if filter_object_id is not None:
        clauses.append('"id_value" = ?')
        all_values.append(str(filter_object_id))  # It's a string in the db schema
    if filter_subject is not None:
        version_ids: list[tuple[int, str, str]] = _VERSION_MANAGER.get_all_versions_subject_id(filter_subject)
        clauses.append(" OR ".join(map(lambda _: f'("log_version" = ? AND "table" = ? AND "field" = ?)', version_ids)))
        all_values.extend(flat_map(lambda tup: list(tup), version_ids))
    elif filter_object_class is not None:
        version_ids: list[tuple[int, str]] = _VERSION_MANAGER.get_all_versions_object_id(filter_object_class)
        clauses.append(" OR ".join(map(lambda _: f'("log_version" = ? AND "table" = ?)', version_ids)))
        all_values.extend(flat_map(lambda tup: list(tup), version_ids))

    where_clause = ""
    if len(clauses) > 0:
        where_clause = f"WHERE ({') AND ('.join(clauses)})"
    
    set_row_count, rows = db_pool.execute_read_statements_in_transaction(
        (f"""
        SELECT COUNT(*) AS "count" FROM "changelog"
        {where_clause}
        """, all_values),
        (f"""
        SELECT *
        FROM "changelog"
        {where_clause}
        ORDER BY "when" DESC, "id" DESC
        LIMIT {entries_per_page} OFFSET {entries_per_page * page}
        """, all_values)
    )
    return (math.ceil(set_row_count[0]["count"]/entries_per_page),
            list(map(_changelog_entry_db_to_json, rows)))


class ChangelogUpdate:
    
    def __init__(self):
        super().__init__()
        self._entries: list[ChangelogEntry] = []
    
    def add_entry(self, entry: ChangelogEntry):
        self._entries.append(entry)
    
    # noinspection PyBroadException
    def _queue_update_statements(self,
                                 transaction: WriteTransaction,
                                 modifying_user_id: int,
                                 change_time: datetime):
        if len(self._entries) == 0:
            return
        # version, table, object id, field, value_old, value_new
        log_list: list[tuple[int, str, int, str, str or None, str]] = []
        for entry in self._entries:
            version = _LEGACY_VERSION
            try:
                if not _VERSION_MANAGER.can_encode_for(entry.subject, version):
                    raise ChangelogEntryError()
                object_id, param1, param2 = _VERSION_MANAGER.encode_entry(version, entry)
            except ChangelogEntryError:
                version = _VERSION_MANAGER.get_current_version_id()
                if not _VERSION_MANAGER.can_encode_for(entry.subject, version):
                    raise RuntimeError("Cannot encode entry for current version")  # pragma: no cover
                object_id, param1, param2 = _VERSION_MANAGER.encode_entry(version, entry)
            
            id = _VERSION_MANAGER.get_subject_id(entry.subject, version)
            table, field = id
            log_list.append((
                version,
                table,
                object_id,
                field,
                param1,
                param2
            ))
        
        transaction.queue_statement(f"""
            INSERT INTO "changelog"
                ("when", "who", "log_version", "table", "id_key", "id_value", "field", "value_old", "value_new", "executed")
                VALUES {",".join(map(lambda _: f"(?, ?, ?, ?, 'id', ?, ?, ?, ?, 1)",
                                     log_list))}
            """, *[v for version, table, object_id, field_, old_value, new_value in log_list
                   for v in  # Flat map a list of lists
                   [change_time,
                    modifying_user_id,
                    version,
                    table,
                    object_id,
                    field_,
                    old_value,
                    new_value]]
                                    )


class UpdateChangelogUpdate(ChangelogUpdate):
    
    def queue_log_statements(self, transaction: WriteTransaction, modifying_user_id: int):
        self._queue_update_statements(transaction, modifying_user_id, datetime.now())


class CreationChangelogUpdate(ChangelogUpdate):
    
    def __init__(self):
        super().__init__()
        self._creation_entries: list[Callable[[int], ChangelogEntry]] = []
    
    def add_object_creation_entry(self, entry_creator: Callable[[int], ChangelogEntry]):
        self._creation_entries.append(entry_creator)
    
    def queue_log_statements(self,
                             transaction: WriteTransaction,
                             modifying_user_id: int,
                             object_id: int):
        change_time = datetime.now()
        
        for entry_creator in self._creation_entries:
            self.add_entry(entry_creator(object_id))
        
        self._queue_update_statements(transaction, modifying_user_id, change_time)
